package watcher

import (
	"fmt"
	"testing"
	"time"

	"git.autistici.org/ale/autoradio/coordination/etcdtest"
)

type testSyncable struct {
	values  map[string]string
	indexes map[string]uint64
}

func (s *testSyncable) Set(key, value string, index uint64) {
	s.values[key] = value
	s.indexes[key] = index
}

func (s *testSyncable) Delete(key string) {
	delete(s.values, key)
	delete(s.indexes, key)
}

func newTestSyncable() *testSyncable {
	return &testSyncable{
		values:  make(map[string]string),
		indexes: make(map[string]uint64),
	}
}

func TestSyncer_IndexCorrectness(t *testing.T) {
	etcd := etcdtest.NewClient()
	target := newTestSyncable()
	syn := NewSyncer(etcd, "/test/", target)
	defer syn.Close()

	n := 1000
	expectedIndex := make(map[string]uint64)
	for i := 0; i < n; i++ {
		key := fmt.Sprintf("/test/key%06d", i)
		resp, _ := etcd.Set(key, fmt.Sprintf("%08x", i), 0)
		expectedIndex[key] = resp.Node.ModifiedIndex
	}

	time.Sleep(100 * time.Millisecond)
	if len(target.values) != n {
		t.Fatalf("Syncable only saw %d keys", len(target.values))
	}
	for key, expected := range expectedIndex {
		idx := target.indexes[key]
		if idx != expected {
			t.Errorf("Synced key has bad index: got %d, expected %d", idx, expected)
		}
	}
}
